/*
A GLUT program using OpenGL, GLUI, and GLSL to do simple volume rendering.

Orion Sky Lawlor, olawlor@acm.org, 2007/02/06 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <fstream>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=2.0; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

float brightness=1.0, shininess=1.0;

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}


/* Draw a big green ground polygon */
void draw_ground(void) {
	double ground=100.0; /* extent of ground in X and Z */
	double groundY=0.0;
	glColor4f(0.1,0.4,0.1,1); /* green */
	glNormal3f(0,1,0); /* facing straight up */
	oglTextureQuad(
		oglVector3d(-ground,groundY,-ground), oglVector3d(+ground,groundY,-ground), 
		oglVector3d(-ground,groundY,+ground), oglVector3d(+ground,groundY,+ground)
	);
}

int volumeSize=128; /* Resolution of volume dataset */
float exposure=0.3; /* exposure scaling */
int argc; char **argv;

int sliceAxisAlong=2; /* <- controls slice axis--slice along Z by default */
vec3 sliceAxis(vec3 v) {
	switch (sliceAxisAlong) {
	case 0: return vec3(v.z,v.x,v.y); /* along x */
	case 1: return vec3(v.x,v.z,v.y); /* along y */
	case 2: return vec3(v.x,v.y,v.z); /* along z */
	}
	return vec3(0.0);
}

void draw_volume(const vec3 &volumeCenter,const vec3 &cameraLoc) 
{
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA); /* premultiplied alpha */
	static oglTextureName *volume=0;
	static int w=volumeSize, h=volumeSize, d=volumeSize; /* dimensions of volume texture in x,y,z */
	GLenum t=GL_TEXTURE_3D;
	if (volume==0) {
		/* Read a 3D volume texture from volume.siz and volume.raw.
		  You can generate these files from a tall skinny JPEG of slices with "jpeg2raw"
		*/
		std::ifstream siz("volume.siz");
		std::string header;
		siz>>header>>w>>h>>d;
		unsigned char *arr=new unsigned char[4*w*h*d];
		printf("Reading %d x %d x %d image from volume.raw...\n",w,h,d);
		std::ifstream raw("volume.raw",std::ios_base::binary);
		raw.read((char *)arr,4*w*h*d);
		volume=new oglTextureName;
		volume->bind(t);
		glTexImage3D(t,0,GL_RGBA8,
			w,h,d,
			0, GL_RGBA, GL_UNSIGNED_BYTE,
			&arr[0]);
		oglTextureType(oglTexture_linear,t);
		glTexParameteri(t, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
		glTexParameteri(t, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
		glTexParameteri(t, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_BORDER);
		delete arr;
	}
	
/* This is GLSL code, that runs on the graphics card itself */
	static oglProgramObject *p=oglProgramFromStrings(
	"/* GLSL Vertex shader, to compute point color and size */\n"
	"uniform vec4 cameraLoc; // Location of camera\n"
	"uniform vec4 sliceStep; // normal of texture slice planes, scaled by distance between slices\n"
	"varying vec4 color; // Material diffuse color\n"
	"varying vec3 texCoord, texStep; // 3D texture coordinates, change per slice\n"
	"void main(void) {\n"
	"	gl_Position=gl_ModelViewProjectionMatrix*gl_Vertex;\n"
	"	color=gl_Color;\n"
	"	texCoord=vec3(gl_Vertex);\n"
	"	vec3 camDir=normalize(vec3(gl_Vertex-cameraLoc));\n"
	"	vec3 sliceDir=normalize(vec3(sliceStep));\n"
	"	texStep=camDir*length(vec3(sliceStep))/dot(camDir,sliceDir);\n"
	"}\n"
	,
	"/* GLSL Fragment shader, to write actual pixels */\n"
	"varying vec4 color;\n"
	"varying vec3 texCoord, texStep;\n"
	"uniform sampler3D volTex;\n"
	"uniform float exposure;\n"
	"void main(void) {\n"
	"	gl_FragColor=color*exposure*texture3D(volTex,texCoord); \n"
	"}\n"
	);
	
	/* Number of slices of the dataset to render
	  Note: the GPU can avoid trilerps and use 2.5x-faster bilerps 
	   *if* the slices cut exactly through voxel planes.
	*/
	int nSlices=volumeSize;
	double sliceSize=1.0/nSlices;
	
	p->begin();
	p->set("cameraLoc",vec4(vec3(cameraLoc),1.0));
	p->set("sliceStep",vec4(sliceAxis(vec3(0,0,sliceSize)),1.0));
	p->set("exposure",brightness/nSlices);
	volume->bind(t);
	p->set("volTex",0); /* volume texture is bound to texture unit 0 */
	
	glDisable(GL_ALPHA_TEST); /* too agressive--loses too many points */
	glDisable(GL_DEPTH_TEST); /* don't do Z buffer (screws up overlaps, esp. w/blending) */
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_3D);
	
	/* Draw the volume dataset as a set of slices */
	glBegin(GL_QUADS);
	double intensity=1.0;
	glColor4f(intensity,intensity,intensity,intensity);
	for (int slice=0;slice<nSlices;slice++) {
		float z=(slice+0.5f)/(float)nSlices;
		float spare=0.0; /* <- extra space to add along all axes */
		glVertex3fv(sliceAxis(vec3(0-spare, 0-spare, z)));
		glVertex3fv(sliceAxis(vec3(1+spare, 0-spare, z)));
		glVertex3fv(sliceAxis(vec3(1+spare, 1+spare, z)));
		glVertex3fv(sliceAxis(vec3(0-spare, 1+spare, z)));
	}
	glEnd();
	
	glDisable(GL_TEXTURE_3D);
	
	p->end();
}


/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	//glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClearColor(0,0,0,0); /* black */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* World location of camera: just back out all the GL_PROJECTION matrix transforms */
	vec3 cameraLoc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	
	// Draw the world 
	glMultMatrixf(&modelMat[0][0]);
	vec3 volumeCenter(0.5,0.5,0.5); /* center point of volume dataset */
	glTranslatefv(-volumeCenter);
	
	draw_volume(volumeCenter,cameraLoc);
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc_, char** argv_)
{
	glutInitWindowSize(800,600);
	glutInit(&argc_,argv_);
	argc=argc_; argv=argv_; /* copy out arguments to global variables */

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	distance=2;
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Distance:", &distance))->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Brightness:", &brightness))->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Shininess:", &shininess))->set_float_limits(0.0, 1.0);
		
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
		new GLUI_Column(glui,false);
		new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
